﻿using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using GASS.CUDA.Types;

namespace Visualizator.Physics
{
    /// <summary>
    /// Representacion del sistema masa-resorte, usando una representacion
    /// de datos ajustada al modelo de integarcion Verlet, es decir, en vez
    /// de almacenar la velocidad actual de una particula, almacenamos su
    /// posicion anterior, de donde sacamos la velocidad.
    /// 
    /// Vamos a necesitar 3 algoritmos.
    /// 1. Acumulacion de fuerzas 
    ///   -> F = m * A.
    ///   -> Aplicar gravedad.
    /// 2. Integracion Verlet (utilizanado: pos, prevPos)
    ///   -> X = 2X - X' + A * T^2
    /// 3. Satisfaccion de resortes (utilizando: pos, mass, springIndex1/2, restLength, k)
    ///   -> F = -k * X
    /// 4. [Opcional] Angular springs?
    /// </summary>
    public class MassSpringSystem
    {
        public Float4 Gravity;

        // Posiciones actuales
        public Float4[] Pos;

        // Posiciones anteriores.
        public Float4[] PrevPos;

        // Fuerzas acumuladas.
        public Float4[] Force;


        /// <summary>
        /// Mas comodo que accesar los float[] directamente.
        /// </summary>
        public void SetPos(int index, Float4 value)
        {
            Pos[index] = value;
        }

        /// <summary>
        /// Mas comodo que accesar los float[] directamente.
        /// </summary>
        public Float4 GetPos(int index)
        {
            return Pos[index];
        }

        /// <summary>
        /// Mas comodo que accesar los float[] directamente.
        /// </summary>
        public void SetPrevPos(int index, Float4 value)
        {
            PrevPos[index] = value;
        }

        /// <summary>
        /// Mas comodo que accesar los float[] directamente.
        /// TODO: Deberia ser ApplyImpulse?
        /// </summary>
        public void SetForce(int index, Float4 value)
        {
            Force[index] = value;
        }

        /// <summary>
        /// La masa de cada punto.
        /// TODO: Sera mejor guardar el inverso de la masa?
        /// </summary>
        public float[] Mass;

        /// <summary>
        /// Inicializa la estructura de datos para almacenar el sistema.
        /// </summary>
        public MassSpringSystem(int particleCount, int springCount)
        {
            Pos = new Float4[particleCount];
            PrevPos = new Float4[particleCount];
            Force = new Float4[particleCount];

            Mass = new float[particleCount];

            SpringIndex1 = new int[springCount];
            SpringIndex2 = new int[springCount];
            RestLength = new float[springCount];
            K = new float[springCount];
        }


        /// <summary>
        /// El indice 1 de los resortes.
        /// </summary>
        public int[] SpringIndex1;

        /// <summary>
        /// El indice 2 de los resortes.
        /// </summary>
        public int[] SpringIndex2;

        /// <summary>
        /// Longitud de descanso del resorte.
        /// </summary>
        public float[] RestLength;

        /// <summary>
        /// Constantes de elasticidad de cada resorte.
        /// </summary>
        public float[] K;

    }
}
