#include "GPUSimulator.h"
#include "Util.h"
#include <iostream>
#include <vector>
#include "cuda_d3d10_interop.h"
#include "DistanceConstraint.h"

using namespace std;
GPUSimulator::GPUSimulator(ID3D10Device *device)
{
	this->device = device;
}

GPUSimulator::~GPUSimulator(void)
{
	this->Dispose();
}

HRESULT GPUSimulator::SetModel( Model<Vector4> *model )
{
	// Los tamaos de los arrays de resortes.
	int *arraySizes;

	// Un arreglo que contiene todos los arreglos de resortes.
	ID3D10Buffer **springArrays;

	this->model = model;

	// Get the info from the model.
	Vector4 *positions =	model->GetVertices();
	int *indices =			model->GetIndices();
	int particleCount =		model->GetVertexCount();

	// Creamos el arreglo de posiciones previas, lo inicializamos
	// con los mismos valores del arreglo de posiciones para que 
	// las particulas tengan velocidad 0 al iniciar la simulacion.
	Vector4 *prevPositions = (Vector4 *)malloc(particleCount * sizeof(Vector4));
	for (int i = 0 ; i < particleCount; i++)
	{
		prevPositions[i] = positions[i];
		prevPositions[i].w = 0;
	}

	// Springs, reservamos la memoria necesaria.
	IntFloat4 *springs = new IntFloat4[(particleCount * (particleCount-1)/2)];

	// Creamos los springs todos contra todos.
	this->springCount = 0;
	for (int i = 0; i < particleCount-1; i++)
	{
		for(int j = i+1; j < particleCount; j++)
		{
			if (Util::Distance(positions[i], positions[j]) > 0.3f) continue;
			if (Util::Distance(positions[i], positions[j]) < 0.001f) continue;

			// Tomamos el valor la constante de elasticidad de las particulas.
			float k = (positions[i].w + positions[j].w) * 0.5f;

			springs[springCount].i.x = i;
			springs[springCount].i.y = j;

			// Copiamos los vectores para no considerar el W en el calculo de distancia.
			Vector4 x1 = positions[i];
			Vector4 x2 = positions[j];
			x1.w = x2.w = 0;
			springs[springCount].f.z = Util::Distance(x1, x2) * 1;
			springs[springCount].f.w = k;

			//cout << springs[k].i.x << " " << (int)springs[k].i.y << " " << springs[k].f.z << " " << springs[k].f.w << endl;
			this->springCount++;
		}
	}

	// Ya no necesitamos la constante de elasticidad por particula.
	for (int i = 0; i < model->GetVertexCount() ; i++)
		positions[i].w = 0;

	// Debugging.
	cout << "Cantidad de particulas: " << particleCount << endl;
	cout << "Cantidad de resortes: " << this->springCount << endl;

	// Ahora tenemos que distribuir los resorte en conjuntos, tal que por cada
	// particula solo haya un resorte que la afecte.

	// Un arreglo de bool para marcar las particulas que ya 
	// tengamos en este array de springs. Y otro para marcar 
	// los springs que ya estan en algun set.
	bool *particlesAlreadyInSet = (bool*) malloc(sizeof(bool) * particleCount);
	bool *springsAlreadyInSomeSet = (bool*) malloc(sizeof(bool) * springCount);
	memset(springsAlreadyInSomeSet, 0, springCount * sizeof(bool));

	// Aqui almacenaremos los conjuntos de springs.
	std::vector<vector<IntFloat4>> springSets = vector<vector<IntFloat4>>();
	bool working = true;
	while (working)
	{
		working = false;
		memset(particlesAlreadyInSet, 0, particleCount * sizeof(bool));
		vector<IntFloat4> set = vector<IntFloat4>();

		for (int i = 0; i < springCount; i++)
		{
			// Si el resorte ya esta en algun set, lo ignoramos.
			if (springsAlreadyInSomeSet[i]) continue;

			IntFloat4 spring = springs[i];
			int particleIndex1 = springs[i].i.x;	// Indices de las particulas a 
			int particleIndex2 = springs[i].i.y;	// cada lado del resorte

			// Si ya hay un resorte que afecte alguna de estas particulas, continuamos.
			if (particlesAlreadyInSet[particleIndex1] || particlesAlreadyInSet[particleIndex2])
			{
				continue;
			}
			particlesAlreadyInSet[particleIndex1] = true;
			particlesAlreadyInSet[particleIndex2] = true;

			working = true;
			springsAlreadyInSomeSet[i] = true;
			// Metemos el resorte en el set actual.
			set.push_back(spring);
		}

		// Guardamos este conjunto de resortes.
		springSets.push_back(set);
	}

	free(particlesAlreadyInSet);
	free(springsAlreadyInSomeSet);
	delete springs;

	// HACK: Al final habremos metido un conjunto extra (vacio), lo quitamos.
	springSets.pop_back();

	// Revisamos que efectivamente la cantidad de resortes se mantuvo
	// luego de distribuiros en los sets.
	int newSpringCount = 0;
	for (UINT i = 0 ; i < springSets.size() ; i++)
	{
		newSpringCount += springSets[i].size();
	}
	AssertOrQuit(newSpringCount == springCount)

		// Creamos el arreglo de arraySizes, aqui almacenaremos el tamao
		// de cada set de spring que tengamos para usarlo en el kernel.
		arraySizes = (int *) malloc(sizeof(int) * springSets.size());
	springArrays = (ID3D10Buffer **) malloc(sizeof(ID3D10Buffer *) * springSets.size());


	D3D10_SUBRESOURCE_DATA posResData, prevPosResData, indexResData, arraySizesResData;
	posResData.pSysMem = positions;
	posResData.SysMemPitch = 0;
	posResData.SysMemSlicePitch = 0;

	prevPosResData.pSysMem = prevPositions;
	prevPosResData.SysMemPitch = 0;
	prevPosResData.SysMemSlicePitch = 0;

	indexResData.pSysMem = indices;
	indexResData.SysMemPitch = 0;
	indexResData.SysMemSlicePitch = 0;

	arraySizesResData.pSysMem = arraySizes;
	arraySizesResData.SysMemPitch = 0;
	arraySizesResData.SysMemSlicePitch = 0;

	// Setup buffer desc
	D3D10_BUFFER_DESC bufferDescParticles;
	bufferDescParticles.Usage            = D3D10_USAGE_DEFAULT;
	bufferDescParticles.ByteWidth        = 16 * particleCount;
	bufferDescParticles.BindFlags        = D3D10_BIND_VERTEX_BUFFER;
	bufferDescParticles.CPUAccessFlags   = 0;
	bufferDescParticles.MiscFlags        = 0;

	D3D10_BUFFER_DESC bufferDescSprings;
	bufferDescSprings.Usage            = D3D10_USAGE_DEFAULT;
	//bufferDescIndices.ByteWidth      = No ponemos ByteWidth porque cada set lleva uno distinto.
	bufferDescSprings.BindFlags        = D3D10_BIND_VERTEX_BUFFER;
	bufferDescSprings.CPUAccessFlags   = 0;
	bufferDescSprings.MiscFlags        = 0;

	D3D10_BUFFER_DESC bufferDescIndices;
	bufferDescIndices.Usage            = D3D10_USAGE_DEFAULT;
	bufferDescIndices.ByteWidth        = sizeof(unsigned int) * model->GetIndexCount();
	bufferDescIndices.BindFlags        = D3D10_BIND_INDEX_BUFFER;
	bufferDescIndices.CPUAccessFlags   = 0;
	bufferDescIndices.MiscFlags        = 0;

	D3D10_BUFFER_DESC bufferDescArraySizes;
	bufferDescArraySizes.Usage            = D3D10_USAGE_DEFAULT;
	bufferDescArraySizes.ByteWidth        = (UINT)sizeof(unsigned int) * springSets.size();
	bufferDescArraySizes.BindFlags        = D3D10_BIND_VERTEX_BUFFER;
	bufferDescArraySizes.CPUAccessFlags   = 0;
	bufferDescArraySizes.MiscFlags        = 0;



	HRESULT hr;
	largestArraySize = 0;
	// Creamos los buffers de resortes
	for (unsigned int i = 0; i < springSets.size(); i++)
	{
		D3D10_SUBRESOURCE_DATA springResData;
		ID3D10Buffer *springBuffer;
		vector<IntFloat4> set = springSets[i];
		IntFloat4 *bufferData = new IntFloat4[set.size()];

		// Almacenamos el tamano de este set de resortes.
		arraySizes[i] = set.size();

		// Copiamos los datos a un array.
		for (int j = 0; j < set.size(); j++)
		{
			bufferData[j] = set[j];
		}

		// Keep the size of the larges array.
		largestArraySize = set.size() > largestArraySize ? set.size() : largestArraySize;

		// Creamos el buffer en memoria de video.
		springResData.pSysMem = bufferData;
		springResData.SysMemPitch = 0;
		springResData.SysMemSlicePitch = 0;
		bufferDescSprings.ByteWidth = 16 * set.size();
		if (FAILED( hr = device->CreateBuffer( &bufferDescSprings, &springResData, &springBuffer )))
			return hr;

		// Guardamos los springs en un array en memoria de CPU. (para el metodo unbatched)
		dSprings.push_back(deviceBuffer(springBuffer, set.size()));

		// Guardamos los springs en un array en memoria de GPU. (para el metodo batched)
		springArrays[i] = springBuffer;
		delete bufferData;
	}

	// Creamos los buffers de las particulas
	if (FAILED( hr = device->CreateBuffer( &bufferDescParticles, &posResData, &dPositions )))
		return hr;
	if (FAILED( hr = device->CreateBuffer( &bufferDescParticles, &prevPosResData, &dPrevPositions )))
		return hr;

	// Creamos el buffer de indices para dibujar los triangulos.
	if (FAILED(hr = device->CreateBuffer( &bufferDescIndices, &indexResData, &pIndices )))
		return hr;

	// Creamos el buffer donde almacenamos los tamaos de cada set de resortes.
	if (FAILED(hr = device->CreateBuffer( &bufferDescArraySizes, &arraySizesResData, &dArraySizes )))
		return hr;


	delete[] positions;
	delete[] prevPositions;

	InitCuda();

	return S_OK;
}

ID3D10Buffer * GPUSimulator::GetVertexBuffer()
{
	return dPositions;
}

void GPUSimulator::InitCuda()
{
	// Comenzamos la interoperabilidad CUDA/DX
	cudaD3D10SetDirect3DDevice(device);
	cudaCheckMsg("cudaD3D10SetDirect3DDevice failed");

	// Registramos los VertexBuffers para que puedan ser utilizados desde
	// los kernels de CUDA.
	cudaD3D10RegisterResource(dPositions, cudaD3D10RegisterFlagsNone);
	cudaCheckMsg("cudaD3D10RegisterResource (pPositions) failed");
	cudaD3D10ResourceSetMapFlags(dPositions, cudaD3D10MapFlagsWriteDiscard);
	cudaCheckMsg("cudaD3D10ResourceSetMapFlags (pPositions) failed");

	cudaD3D10RegisterResource(dPrevPositions, cudaD3D10RegisterFlagsNone);
	cudaCheckMsg("cudaD3D10RegisterResource (pPrevPositions) failed");
	cudaD3D10ResourceSetMapFlags(dPrevPositions, cudaD3D10MapFlagsWriteDiscard);
	cudaCheckMsg("cudaD3D10ResourceSetMapFlags (pPrevPositions) failed");

	cudaD3D10RegisterResource(dArraySizes, cudaD3D10RegisterFlagsNone);
	cudaCheckMsg("cudaD3D10RegisterResource (pArraySizes) failed");
	cudaD3D10ResourceSetMapFlags(dArraySizes, cudaD3D10MapFlagsWriteDiscard);
	cudaCheckMsg("cudaD3D10ResourceSetMapFlags (pArraySizes) failed");


	for (int i = 0; i < dSprings.size() ; i++)
	{
		cudaD3D10RegisterResource(dSprings[i].buffer, cudaD3D10RegisterFlagsNone);
		cudaCheckMsg("cudaD3D10RegisterResource (pSprings) failed");
		cudaD3D10ResourceSetMapFlags(dSprings[i].buffer, cudaD3D10MapFlagsWriteDiscard);
		cudaCheckMsg("cudaD3D10ResourceSetMapFlags (pSprings) failed");
	}

	cudaMalloc(&dSpringArrays, sizeof(void *) * dSprings.size());

	// Variable para guardar temporalmente el apuntador de cada
	// spring buffer mapeado.
	void *dSpringMappedPtr;

	//// CUDA Map call to the Vertex Buffer and return a pointer
	cudaD3D10MapResources(1, (ID3D10Resource **)&dPositions);
	//cudaCheckMsg("cudaD3D10MapResource failed");
	cudaD3D10ResourceGetMappedPointer( &dPosMappedPtr, dPositions, 0);
	//cudaCheckMsg("cudaD3D10ResourceGetMappedPointer failed");

	cudaD3D10MapResources(1, (ID3D10Resource **)&dPrevPositions);
	//cudaCheckMsg("cudaD3D10MapResource failed");
	cudaD3D10ResourceGetMappedPointer( &dPrevPosMappedPtr, dPrevPositions, 0);
	//cudaCheckMsg("cudaD3D10ResourceGetMappedPointer failed");

	cudaD3D10MapResources(1, (ID3D10Resource **)&dArraySizes);
	//cudaCheckMsg("cudaD3D10MapResource failed");
	cudaD3D10ResourceGetMappedPointer( &dArraySizesMappedPtr, dArraySizes, 0);
	cudaCheckMsg("cudaD3D10ResourceGetMappedPointer failed");

	void **springArrays = new void*[dSprings.size()];
	// Mapeamos todos los buffers de springs.
	for (int i = 0; i < dSprings.size() ; i++)
	{
		cudaD3D10MapResources(1, (ID3D10Resource **)&dSprings[i].buffer);
		//cudaCheckMsg("cudaD3D10MapResource failed");
		cudaD3D10ResourceGetMappedPointer( &dSpringMappedPtr, dSprings[i].buffer, 0);
		//cudaCheckMsg("cudaD3D10ResourceGetMappedPointer failed");

		springArrays[i] = dSpringMappedPtr;
	}

	// Copiamos los apuntadores a los spring arrays a memoria de video.
	cudaMemcpy(dSpringArrays, springArrays, dSprings.size() * sizeof(void *), cudaMemcpyHostToDevice);

	delete[] springArrays;
}

void GPUSimulator::Step( float dt )
{
	// Ejecutamos el kernel de integracion Verlet.
	ExecuteIntegrate(dPosMappedPtr, dPrevPosMappedPtr, model->GetVertexCount(), dt);

	// Ejecutamos el kernel de resortes.
	DistanceConstraint::ExecuteBatched(dPosMappedPtr, (void **)dSpringArrays, dArraySizesMappedPtr, dSprings.size(), largestArraySize, 1, dt);
	//Draw(dt);
	cudaThreadSynchronize();
}

void GPUSimulator::Dispose()
{
	// Desmapeamos todos los buffers que teniamos registrados
	// con CUDA.
	for (int i = 0; i < dSprings.size() ; i++)
	{
		cudaD3D10UnmapResources(1, (ID3D10Resource **)&dSprings[i].buffer);
	}

	//// CUDA Map Unmap vertex buffer
	cudaD3D10UnmapResources(1, (ID3D10Resource **)&dArraySizes);
	//cudaCheckMsg("cudaD3D10UnmapResources failed");
	cudaD3D10UnmapResources(1, (ID3D10Resource **)&dPositions);
	//cudaCheckMsg("cudaD3D10UnmapResources failed");
	cudaD3D10UnmapResources(1, (ID3D10Resource **)&dPrevPositions);
	//cudaCheckMsg("cudaD3D10UnmapResources failed");
	
	// Liberamos la memoria en video.
	dPositions->Release();
	dPrevPositions->Release();
	for (int i = 0; i < dSprings.size() ; i++)
	{
		dSprings[i].buffer->Release();
	}
}
